'ChirpPix.bas 2.Apr 03
'by Markus Vester, DF6NM

'history:
'bmftwa2d 15.12.99, 2f 18.06.00, 2g 20.12.02, 2h 6.3.03
'2j 10.3.03 cos^2-windowed fft
'2n 2.4.03 reads 8 and 24 bit bmps

'$DYNAMIC
 DEFINT I-N

 CLEAR
 PRINT "Chirped Fourier Picture Transmission by DF6NM"

'set defaults and read cmdline parameters
 lm = 13    'max FFT length 8192
 f1! = 2160  'Hz
 path$ = ""
 'path$ = "d:\vester\basic\fft\bmp_FFT\images\"
 'path$ = "c:\eigene~1\markus\basic\fft\bmp_FF~1\images\"
 file$ = "image"
 samplerate! = 6245.641 'Hz
  ' = 0.76241Hz/p * 1.312s/p  (128x87 = 97.6Hz x 114.1s)
  ' Argo 3s slow (0.337Hz/p 0.600s/p) gives
  ' image scaling 2.266*height 2.186*width
  ' /sr 6135 would be isotropic (aspect ratio 1:1)
 winflag% = 1
 ampl! = 1
 chirp! = 1

 comm$ = COMMAND$
 IF INSTR(comm$, "?") > 0 OR INSTR(comm$, "/help") > 0 THEN
  PRINT "command line parameters and defaults:"
  PRINT "/file image.bmp /fl 2160 /sr 6245.641 /fft 13 /ampl 1 /chirp 1 /nowin"
  GOTO ende
 END IF
 argp = INSTR(comm$, "/SR"): IF argp > 0 THEN samplerate! = VAL(MID$(comm$, argp + 3))
 argp = INSTR(comm$, "/FL"): IF argp > 0 THEN f1 = VAL(MID$(comm$, argp + 3))
 argp = INSTR(comm$, "/FFT"): IF argp > 0 THEN lm = VAL(MID$(comm$, argp + 4))
 IF lm > 13 THEN lm = 13
 IF lm < 4 THEN lm = 4
 argp = INSTR(comm$, "/AMPL"): IF argp > 0 THEN ampl! = VAL(MID$(comm$, argp + 5))
 argp = INSTR(comm$, "/CHIRP"): IF argp > 0 THEN chirp! = VAL(MID$(comm$, argp + 6))
 argp = INSTR(comm$, "/NOWIN"): IF argp > 0 THEN winflag% = 0
 argp = INSTR(comm$, "/FILE "): argp2 = INSTR(comm$, ".BMP")
 IF argp > 0 AND argp2 > argp + 6 THEN file$ = MID$(comm$, argp + 6, argp2 - argp - 6)

 file1$ = path$ + file$ + ".bmp"
 file2$ = path$ + file$ + ".wav"

'prepare FFT
 nm = 2 ^ lm: nm1 = nm - 1
 DIM a!(2 * nm - 1), w!(nm1), nbr%(nm1)
 DIM mc%(155) 'for 312 bytes
 DIM col!(255)
 out$ = STRING$(nm, 128)

 GOSUB getmc
 GOSUB calcw
 GOSUB calcbitrev
 IF winflag% THEN
  DIM win!(nm1), aold!(nm1)
  GOSUB calcwin
 END IF

 OPEN file1$ FOR INPUT AS 1: CLOSE 1 'check if bmp file exists
 OPEN file1$ FOR BINARY AS 1
 OPEN file2$ FOR OUTPUT AS 2

'read 54 byte bmp-header
 i$ = INPUT$(2, 1)
 IF i$ <> "BM" THEN PRINT "input file not in bmp format": GOTO closefiles
 i$ = INPUT$(8, 1)
 hdrlen = CVL(INPUT$(4, 1))
 i$ = INPUT$(4, 1)
 xm = CVL(INPUT$(4, 1))
 ym = CVL(INPUT$(4, 1))
 i$ = INPUT$(2, 1)
 depth% = CVI(INPUT$(2, 1))
 i$ = INPUT$(24, 1)

 IF depth% <> 8 AND depth% <> 24 THEN PRINT "256 or 16M colours required": GOTO closefiles
 bw! = .11: gw! = .59: rw! = .3 'greyscale colour weights

'read colourmap
 IF depth% <= 8 THEN
  numcols = 2 ^ depth%
  FOR n = 0 TO numcols - 1
   i$ = INPUT$(4, 1)
   col!(n) = bw! * ASC(MID$(i$, 1, 1)) + gw! * ASC(MID$(i$, 2, 1)) + rw! * ASC(MID$(i$, 3, 1))
  NEXT n
 END IF
'bytes per line in bmp
 IF depth% = 8 THEN xm4 = 4 * ((xm + 3) \ 4)
 IF depth% = 24 THEN xm4 = 4 * ((xm * 3 + 3) \ 4)
 ymm = ym + winflag%

'frequency band
 nf1 = f1 / samplerate! * nm
 f1act = nf1 * samplerate! / nm
 f2act = (nf1 + xm - 1) * samplerate! / nm
 PRINT file$ + ".bmp "; xm; "x"; ym;
 PRINT f1act; "-"; f2act; "Hz "; samplerate!; "samples/s"
 PRINT samplerate! / nm; "Hz/pix "; nm / samplerate!; "s/pix (";
 IF winflag% = 0 THEN PRINT "un";
 PRINT "smoothed)"

'calculate quadratic phasefactors:
 DIM phr!(xm - 1), phi!(xm - 1)
 a0! = ampl! / 2 / SQR(xm * 2)
 FOR n = 0 TO xm - 1
  ph! = chirp! * (-pi! / xm) * n * n
  phr!(n) = a0! * COS(ph!): phi!(n) = a0! * SIN(ph!)
 NEXT n
  
'write wav header (8 bit mono)
 samples = ymm * nm
 samples2 = 2 * ((samples + 1) \ 2)
 bits = 8
 channels = 1
 bytespersample = bits \ 8 * channels
 file2len = 48& + samples2 * bytespersample
 h2$ = h2$ + "RIFF" + MKL$(file2len - 8)
 h2$ = h2$ + "WAVEfmt " + MKL$(16)
 h2$ = h2$ + MKI$(1) + MKI$(channels)
 h2$ = h2$ + MKL$(samplerate!)
 h2$ = h2$ + MKL$(samplerate! * bytespersample)
 h2$ = h2$ + MKI$(bytespersample) + MKI$(bits)
 h2$ = h2$ + "data" + MKL$(samples)
 PRINT #2, h2$;

'Fourier transform lines and write audio data
 FOR y = 0 TO ymm - 1
  FOR n = 0 TO 2 * nm - 1
   a!(n) = 0
  NEXT n

  IF y < ym THEN
   LOCATE CSRLIN, 1: PRINT y + 1;
   i$ = INPUT$(xm4, 1)
   FOR n = 0 TO xm - 1
    nn2 = ((n + nf1) AND nm1) * 2
    IF depth% = 24 THEN
     n1 = (xm - 1 - n) * 3 + 1
     ap! = bw! * ASC(MID$(i$, n1, 1)) + gw! * ASC(MID$(i$, n1 + 1, 1)) + rw! * ASC(MID$(i$, n1 + 2, 1))
    ELSE
     ap! = col!(ASC(MID$(i$, xm - n, 1)))
    END IF
    a!(nn2) = ap! * phr!(n): a!(nn2 + 1) = ap! * phi!(n)
   NEXT n
   GOSUB fftmc
  END IF
 
  nm21 = 2 * nm - 1
  FOR n = 0 TO nm - 1
   IF winflag% THEN
    n2 = (2 * n + nm) AND nm21
    o! = a!(n2) * win!(n) + aold!(n) * win!(nm - 1 - n) + 128
    aold!(n) = a!(n2)
   ELSE
    o! = a!(n * 2) + 128
   END IF
   IF o! > 255 THEN o! = 255
   IF o! < 0 THEN o! = 0
   MID$(out$, n + 1, 1) = CHR$(o!)
  NEXT n
  PRINT #2, out$;

  IF INKEY$ = CHR$(27) THEN GOTO closefiles

 NEXT y

 LOCATE CSRLIN, 1: PRINT file$ + ".wav ready"

closefiles:
 CLOSE 2
 CLOSE 1
ende:
END

'****************
'FFT stuff

fftmc:
 DEF SEG = VARSEG(mc%(0))
 CALL absolute(0)
RETURN

calcw:
'nm/2 warp factors
  sc! = 1
  pi! = 4 * ATN(1)
  FOR n = 0 TO nm \ 2 - 1
   w!(2 * n) = sc! * COS(2 * pi! * n / nm)
   w!(2 * n + 1) = sc! * SIN(2 * pi! * n / nm)
  NEXT n
 RETURN

calcbitrev:
'bit reversal table
  FOR n = 0 TO nm - 1
   n2 = 0: n1 = n
   FOR l = 0 TO lm - 1
    n2 = 2 * n2 + (n1 AND 1)
    n1 = n1 \ 2
   NEXT l
   nbr%(n) = n2
  NEXT n
 RETURN

calcwin:
'cos^2 window
 FOR n = 0 TO nm - 1
  win!(n) = (1! - COS(pi! * n / nm)) / 2!
 NEXT n
RETURN

getmc:
'get FFT code
  RESTORE float4a
  DEF SEG = VARSEG(mc%(0))
  n0 = 0
  READ i$
  WHILE i$ <> "end"
   FOR n = 0 TO LEN(i$) \ 2 - 1
    POKE n0 + n, VAL("&h" + MID$(i$, 2 * n + 1, 2))
   NEXT n
   n0 = n0 + LEN(i$) \ 2
   READ i$
  WEND
  DEF SEG
  mc%(2) = VARSEG(a!(0))
  mc%(3) = VARSEG(w!(0))
  mc%(4) = VARSEG(nbr%(0))
  mc%(5) = lm
  mc%(6) = nm
 RETURN

float4a:
'assembler FFT by Markus Vester, 1999
DATA E9150090000000000000030008000800
DATA 0100000000000000061E2E8E1E04002E
DATA 8E0608002E8B0E0C00BF0000D1E7268B
DATA 35D1EF3BFE7E35C1E603C1E703FF34FF
DATA 358F048F05FF7402FF75028F44028F45
DATA 02FF7404FF75048F44048F4504FF7406
DATA FF75068F44068F4506C1EF0347E2BD2E
DATA 8E0606009BDBE32EC706100001002EA1
DATA 0C00D1E8C1E0032EA30E002EA10A002E
DATA A316002EC706120000002E8B0E10002E
DATA 8B3E12002E8B36120001CEC1E703C1E6
DATA 03BB000026D90726D94704D904D94404
DATA D9C1D8CCD9C1D8CCDEE9D9CAD8CBD9C9
DATA D8CCDEC1D905D94504D9C0D8E3D95C04
DATA D9C1D8E4D91CD8C2D95D04D8C2D91DDD
DATA C3DDC2DDC1DDC083C60883C7082E031E
DATA 0E00E2B02EA112002E030610002E0306
DATA 10002EA312002E3B060C007403E97AFF
DATA 2ED12610002ED12E0E002EFF0E160074
DATA 03E95FFF1F07CB90
DATA end

